@@ -1,28 +1,36 @@
 #!/usr/bin/env python2
 # vim:fileencoding=utf-8
+# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>
 from __future__ import absolute_import, division, print_function, unicode_literals
 
-__license__ = 'GPL v3'
-__copyright__ = '2015, Kovid Goyal <kovid at kovidgoyal.net>'
-
-import os, subprocess, sys
+import os
+import subprocess
+import sys
 from threading import Thread
 
 from PyQt5.Qt import (
-    QFormLayout, QLineEdit, QToolButton, QHBoxLayout, QLabel, QIcon, QPrinter,
-    QPageSize, QComboBox, QDoubleSpinBox, QCheckBox, QProgressDialog, QTimer)
+    QCheckBox, QComboBox, QDoubleSpinBox, QFormLayout, QHBoxLayout, QIcon, QLabel,
+    QLineEdit, QPageSize, QPrinter, QProgressDialog, QTimer, QToolButton
+)
 
 from calibre import sanitize_file_name
-from calibre.ptempfile import PersistentTemporaryFile
 from calibre.ebooks.conversion.plugins.pdf_output import PAPER_SIZES
-from calibre.gui2 import elided_text, error_dialog, choose_save_file, Application, open_local_file, dynamic
+from calibre.gui2 import (
+    Application, choose_save_file, dynamic, elided_text, error_dialog,
+    open_local_file
+)
 from calibre.gui2.widgets2 import Dialog
-from calibre.gui2.viewer.main import vprefs
+from calibre.ptempfile import PersistentTemporaryFile
+from calibre.utils.config import JSONConfig
+from calibre.utils.filenames import expanduser
 from calibre.utils.icu import numeric_sort_key
 from calibre.utils.ipc.simple_worker import start_pipe_worker
 from calibre.utils.serialize import msgpack_dumps, msgpack_loads
 
 
+vprefs = JSONConfig('viewer')
+
+
 class PrintDialog(Dialog):
 
     OUTPUT_NAME = 'print-to-pdf-choose-file'
@@ -40,7 +48,7 @@ class PrintDialog(Dialog):
         self.file_name = f = QLineEdit(self)
         val = dynamic.get(self.OUTPUT_NAME, None)
         if not val:
-            val = os.path.expanduser('~')
+            val = expanduser('~')
         else:
             val = os.path.dirname(val)
         f.setText(os.path.abspath(os.path.join(val, self.default_file_name)))
@@ -162,6 +170,8 @@ def do_print():
     data = msgpack_loads(stdin.read())
     ext = data['input'].lower().rpartition('.')[-1]
     input_plugin = plugin_for_input_format(ext)
+    if input_plugin is None:
+        raise ValueError('Not a supported file type: {}'.format(ext.upper()))
     args = ['ebook-convert', data['input'], data['output'], '--paper-size', data['paper_size'], '--pdf-add-toc',
             '--disable-remove-fake-margins', '--chapter-mark', 'none', '-vv']
     if input_plugin.is_image_collection:
